unit LithuanianCharSetConverter_unit;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, ComCtrls, StdCtrls,
  IniFiles, Types, LazFileUtils, lazutf8classes, FileUtil, Math, LCLIntf, Menus,

  //  Displays file version info for
  //- Windows PE executables
  //- Linux ELF executables (compiled by Lazarus)
  //- macOS MACH-O executables (compiled by Lazarus)
  //  Runs on Windows, Linux, macOS
  {$IFDEF UNIX}{$IFDEF UseCThreads}
  cthreads,
  {$ENDIF}{$ENDIF}
  // FPC 3.0 fileinfo reads exe resources as long as you register the appropriate units
  fileinfo,
  winpeimagereader, //need this for reading exe info
  elfreader, //needed for reading ELF executables
  machoreader; //needed for reading MACH-O executables


type

  { TLithuanianConverterForm }

  TLithuanianConverterForm = class(TForm)
    AboutMemo: TMemo;
    MainMenu: TMainMenu;
    ResultsMemo: TMemo;
    OpenOutputDirectoryButton: TButton;
    OpenInputDirectoryButton: TButton;
    StartBatchConversionButton: TButton;
    StopBatchConversionButton: TButton;
    FileExtensionsEdit: TEdit;
    IncludeSubdirectoriesCheckBox: TCheckBox;
    ConvertProgressLabel: TLabel;
    FileExtensionsLabel: TLabel;
    OutputDirectoryNameEdit: TEdit;
    InputDirectoryNameEdit: TEdit;
    InputDirectoryNameLabel: TLabel;
    OutputDirectoryNameLabel: TLabel;
    OutputMemo: TMemo;
    OutputFileFormatComboBox: TComboBox;
    InputFileFormatComboBox: TComboBox;
    InputFileFormatLabel: TLabel;
    OutputFileFormatLabel: TLabel;
    OpenOutputFileButton: TButton;
    ConvertSingleFileButton: TButton;
    OpenInputFileButton: TButton;
    OpenOutputFileCheckBox: TCheckBox;
    OutputFileComboBox: TComboBox;
    OutpuFileAutoNameLabel: TLabel;
    OutputFileNameEdit: TEdit;
    InputFileNameEdit: TEdit;
    InputFileNameLabel: TLabel;
    OutputFileNameLabel: TLabel;
    LithuanianCharsetPageControl: TPageControl;
    ConvertProgressBar: TProgressBar;
    TabSheetSingle: TTabSheet;
    TabSheetBatchConvert: TTabSheet;
    TabSheetAbout: TTabSheet;
    procedure ConvertSingleFileButtonClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure IncludeSubdirectoriesCheckBoxChange(Sender: TObject);
    procedure InputFileFormatComboBoxChange(Sender: TObject);
    procedure OpenInputDirectoryButtonClick(Sender: TObject);
    procedure OpenInputFileButtonClick(Sender: TObject);
    procedure OpenOutputDirectoryButtonClick(Sender: TObject);
    procedure OpenOutputFileButtonClick(Sender: TObject);
    procedure OpenOutputFileCheckBoxChange(Sender: TObject);
    procedure OutputFileComboBoxChange(Sender: TObject);
    procedure OutputFileFormatComboBoxChange(Sender: TObject);
    procedure StartBatchConversionButtonClick(Sender: TObject);
    procedure StopBatchConversionButtonClick(Sender: TObject);
  private

  public
    function MessageDlgEx(const AMsg: string; ADlgType: TMsgDlgType; AButtons: TMsgDlgButtons; AParent: TForm): TModalResult;
    procedure CheckIfFileIsUTF8BOM(input_filename_string: TFileName);
    procedure ConvertLithuanianFile(input_filename_string: TFileName; output_filename_string: TFileName);


  end;

var
  LithuanianConverterForm: TLithuanianConverterForm;

  FileVerInfo: TFileVersionInfo;

  input_file_stream: TStream;
  output_file_stream: TStream;
  file_size: int64;

  one_byte: byte;

  input_file_is_utf8_format: boolean;
  n: integer;

  input_file_data: array of byte;
  number_of_replaced_letters: int64;


  input_charset_select: integer;
  output_charset_select: integer;
  autofilenaming_charset_select: integer;
  open_file_after_conversion_select: boolean;
  eFormFiller_enabled: boolean;

  charset_statistics_of_input_file: array of int64;
  charset_probability_table: array[0..7] of int64;

  ProgramIniFile: TIniFile;
  ini_filename, ini_filepath: TFileName;
  exe_filepath, application_path, test_path: TFileName;

  input_directory, old_input_directory, output_directory, old_output_directory,
  output_directory_short, output_filename: TFileName;

  input_file_list: TStringList;
  input_file_formats_stringlist: TStringList;

  file_extensions_string, file_extensions_string_default: string;
  stop_conversion: boolean;

  input_directory_length, output_directory_length: integer;
  include_subdirectories: boolean;

  number_of_converted_files: integer;
  number_of_nulls: integer;


//array size 8x18
const
  LithuanianCharSetTable: array of byte = (
  $41,$43,$45,$45,$49,$53,$55,$55,$5A,$61,$63,$65,$65,$69,$73,$75,$75,$7A, // ASCII
  $B5,$B6,$B7,$B8,$BD,$BE,$C6,$C7,$CF,$D0,$D1,$D2,$D3,$D4,$D5,$D6,$D7,$D8, // 772/774/775
  $8F,$80,$8B,$90,$AD,$9F,$A7,$A6,$92,$85,$87,$8A,$82,$8D,$9E,$97,$96,$91, // 770 (IBM Baltic)
  $DC,$DE,$F0,$F2,$F4,$F6,$F8,$FA,$FC,$DD,$DF,$F1,$F3,$F5,$F7,$F9,$FB,$FD, // 771/773/KBL
  $C0,$C8,$C6,$CB,$C1,$D0,$D8,$DB,$DE,$E0,$E8,$E6,$EB,$E1,$F0,$F8,$FB,$FE, // Windows-1257/Baltic-RIM/Latin-7/ISO-8859-13
  $A1,$C8,$CA,$CC,$C7,$A9,$D9,$DE,$AE,$B1,$E8,$EA,$EC,$E7,$B9,$F9,$FE,$BE, // Latin-4/ISO-8859-4/Latin-6/ISO-8859-10
  $C4,$C4,$C4,$C4,$C4,$C5,$C5,$C5,$C5,$C4,$C4,$C4,$C4,$C4,$C5,$C5,$C5,$C5, // UTF-8 first byte
  $84,$8C,$98,$96,$AE,$A0,$B2,$AA,$BD,$85,$8D,$99,$97,$AF,$A1,$B3,$AB,$BE); // UTF-8 second byte


implementation

{$R *.lfm}

{ TLithuanianConverterForm }



function TLithuanianConverterForm.MessageDlgEx(const AMsg: string; ADlgType: TMsgDlgType; AButtons: TMsgDlgButtons; AParent: TForm): TModalResult;
var
  MsgFrm: TForm;
begin
  MsgFrm := CreateMessageDialog(AMsg, ADlgType, AButtons);
  try
    MsgFrm.Position := poDefaultSizeOnly;
    MsgFrm.FormStyle := fsSystemStayOnTop;
    MsgFrm.Left := AParent.Left + (AParent.Width - MsgFrm.Width) div 2;
    MsgFrm.Top := AParent.Top + (AParent.Height - MsgFrm.Height) div 2;
    Result := MsgFrm.ShowModal;
  finally
    MsgFrm.Free
  end;
end;


procedure TLithuanianConverterForm.CheckIfFileIsUTF8BOM(
  input_filename_string: TFileName);
begin

  if not FileExistsUTF8(input_filename_string) then
    exit;

  input_file_stream := TFileStreamUTF8.Create(input_filename_string, fmOpenRead or fmShareDenyWrite);
  try
    input_file_stream.Position := 0;
    file_size := input_file_stream.Size;
    SetLength(input_file_data, 3 div SizeOf(byte));
    input_file_stream.Read(PByte(input_file_data)^, 3);
  finally
    input_file_stream.Free;
  end;


  //if(((int)input_file_data[0] = $EF)&((int)input_file_data[1] = $BB)&((int)input_file_data[2] = $BF)) then // UTF-8 BOM (byte order mark)
  if ((input_file_data[0] = 239) and (input_file_data[1] = 187) and
    (input_file_data[2] = 191)) then  // UTF-8 BOM (byte order mark)
  begin
    input_file_is_utf8_format := True;
    input_charset_select := 6; // UTF-8 BOM (byte order mark)
  end
  else
  begin
    input_file_is_utf8_format := False;
  end;


  SetLength(input_file_data, 0);

end;


procedure TLithuanianConverterForm.ConvertLithuanianFile(input_filename_string: TFileName;  output_filename_string: TFileName);
var
  i, i2, i3: integer;
  current_value: byte;
  table_value, table_value2: byte;
  current_byte, current_byte2, current_byte3: byte;
  icharset: integer;

  charset_n: integer;
  number_of_charset_letters: integer;

  autodetected_charset: integer;
  max_char_count: int64;
  start_i: int64;

begin

  number_of_replaced_letters := 0;
  number_of_nulls := 0;

  if (LithuanianCharsetPageControl.ActivePage = TabSheetSingle) then
  begin
    OutputMemo.Lines.Clear();
    ResultsMemo.Lines.Clear();
  end;



  if not FileExistsUTF8(input_filename_string) then
  begin
    MessageDlgEx('Error: Unable to open input file:' + LineEnding + string(input_filename_string), mtInformation, [mbOK], LithuanianConverterForm);
    ConvertSingleFileButton.Enabled := True;
    OpenInputFileButton.Enabled := True;
    OpenOutputFileButton.Enabled := True;
    Cursor := crDefault;
    Exit;
  end;

  input_file_stream := TFileStreamUTF8.Create(input_filename_string, fmOpenRead or fmShareDenyWrite);
  try
    input_file_stream.Position := 0;
    file_size := input_file_stream.Size;
    SetLength(input_file_data, file_size div SizeOf(byte));
    input_file_stream.Read(PByte(input_file_data)^, file_size);
  finally
    input_file_stream.Free;
  end;

  input_charset_select := InputFileFormatComboBox.ItemIndex - 1;
  output_charset_select := OutputFileFormatComboBox.ItemIndex;

  autodetected_charset := 0;
  max_char_count := 0;


  //=========================
  if (input_charset_select = -1) then // Autodetect Lithuanian charset
  begin
    //if(((int)input_file_data[0] = $EF) and ((int)input_file_data[1] = $BB) and ((int)input_file_data[2] = $BF)) then // UTF-8 BOM (byte order mark)
    if ((input_file_data[0] = 239) and (input_file_data[1] = 187) and (input_file_data[2] = 191)) then   // UTF-8 BOM (byte order mark)
    begin
      autodetected_charset := 6; // UTF-8 BOM (byte order mark)
    end
    else
    begin
      // not UTF-8 BOM (byte order mark) BEGIN
      for i := 0 to 8 - 1 do
      begin
        charset_probability_table[i] := 0;
      end;

      for i := 0 to 256 - 1 do
      begin
        charset_statistics_of_input_file[i] := 0;
      end;
      //======================

      // Count number of Lithuanian UTF-8 characters
      for i := 0 to file_size - 2 do
      begin
        current_byte := integer(input_file_data[i]);
        current_byte2 := integer(input_file_data[i + 1]);

        //-----------
        if (current_byte = 0) then
        begin
          Inc(number_of_nulls);
        end
        else if (current_byte > 127) then
        begin

          for icharset := 0 to 18 - 1 do
          begin
            table_value := LithuanianCharSetTable[icharset + 108]; // UTF-8 first byte
            table_value2 := LithuanianCharSetTable[icharset + 126]; // UTF-8 second byte
            if ((table_value = current_byte) and (table_value2 = current_byte2)) then
            begin
              Inc(charset_probability_table[7]); // UTF-8
              break;
            end; //if ((table_value = current_byte) and (table_value2 = current_byte2)) then
          end; // for icharset := 0 to 18 - 1 do
        end;
        //else if(current_byte > 127)
        //-----------

      end; // for i := 0 to file_size - 2 do
      //======================

      for i := 0 to file_size - 1 do
      begin
        current_byte := input_file_data[i];
        Inc(charset_statistics_of_input_file[current_byte]);
      end; //for i := 0 to file_size - 1 do

      for i := 128 to 256 - 1 do
      begin
        number_of_charset_letters := charset_statistics_of_input_file[i];

        for charset_n := 0 to 6 - 1 do
        begin

          for icharset := 0 to 18 - 1 do
          begin
            table_value := LithuanianCharSetTable[icharset + (18 * charset_n)];///crash

            if (table_value = i) then
            begin
              charset_probability_table[charset_n] :=
                charset_probability_table[charset_n] + number_of_charset_letters;
              break;
            end;

          end; // for icharset := 0 to 18 - 1 do
        end; // for charset_n := 0 to 6 - 1 do

      end; //for i := 128 to 256 - 1 do



      for i := 0 to 8 - 1 do
      begin
        if (max_char_count < charset_probability_table[i]) then
        begin
          max_char_count := charset_probability_table[i];
          autodetected_charset := i;
        end;
        //if (autodetected_charset < charset_probability_table[i])

      end; //for i := 0 to 8 - 1 do

    end; // not UTF-8 BOM (byte order mark) END



    input_charset_select := autodetected_charset;
    //InputFileFormatComboBoxChange(Self);

    //if (LithuanianCharsetPageControl.ActivePage = TabSheetBatchConvert) then
    //begin

    if (input_charset_select = output_charset_select) then
    begin
      CopyFile(input_filename_string, output_filename_string, [cffOverwriteFile, cffCreateDestDirectory, cffPreserveTime]);
      Exit;
    end;

    //end; //if (LithuanianCharsetPageControl.ActivePage = TabSheetBatchConvert) then


    if (LithuanianCharsetPageControl.ActivePage = TabSheetSingle) then
    begin
      ResultsMemo.Lines.Add('Autodetected charset: ' + InputFileFormatComboBox.Items.Strings[autodetected_charset + 1]);
    end;

  end; // if (input_charset_select = -1) then // Autodetect Lithuanian charset
  //=========================




  current_value := 0;
  icharset := 0;
  table_value := 0;
  table_value2 := 0;
  i := 0;
  i2 := 0;
  i3 := 0;
  n := 0;



  if ((input_charset_select > 5) and (output_charset_select < 6)) then // input UTF-8 and output not UTF-8
  begin

    if FileExistsUTF8(output_filename_string) then
      DeleteFileUTF8(output_filename_string);

    try
      output_file_stream := TFileStreamUTF8.Create(output_filename_string, fmCreate or fmOpenWrite);
      try
        //======================
        start_i := 0; //if((input_charset_select = 7)

        if (input_charset_select = 6) then
        begin
          start_i := 3;
        end;


        // Count number of Lithuanian UTF-8 characters
        for i := start_i to file_size - 1 do
        begin

          current_byte := input_file_data[i];

          i2 := i + 1;
          if(i2 < file_size) then
          current_byte2 := input_file_data[i2]
          else
          current_byte2 := 0;

          i3 := i + 2;
          if(i3 < file_size) then
          current_byte3 := input_file_data[i3]
          else
          current_byte3 := 0;

          //-----------
          if (current_byte = 0) then
          begin
            Inc(number_of_nulls);
          end
          else if (current_byte > 127) then
          begin


            if ((current_byte = $E2) and (current_byte2 = $80)) then
            begin

            if (current_byte3 = $90) then  // Hyphen
            begin
              current_value := $2D;
              output_file_stream.WriteBuffer(current_value, 1);
            end

            else if (current_byte3 = $91) then // Non-Breaking Hyphen
            begin
              current_value := $2D;
              output_file_stream.WriteBuffer(current_value, 1);
            end

            else if (current_byte3 = $92) then // Figure Dash
            begin
              current_value := $2D;
              output_file_stream.WriteBuffer(current_value, 1);
            end

            else if (current_byte3 = $93) then // En Dash
            begin
              current_value := $2D;
              output_file_stream.WriteBuffer(current_value, 1);
            end

            else if (current_byte3 = $94) then // Em Dash
            begin
              current_value := $2D;
              output_file_stream.WriteBuffer(current_value, 1);
            end

            else if (current_byte3 = $95) then // Horizontal Bar
            begin
              current_value := $2D;
              output_file_stream.WriteBuffer(current_value, 1);
            end

            end // if(current_byte = $E2) and (current_byte2 = $80) then


            else if (current_byte = $EF) then
            begin

            if ((current_byte2 = $B9) and (current_byte3 = $98)) then  // Small Em Dash
            begin
              current_value := $2D;
              output_file_stream.WriteBuffer(current_value, 1);
            end

            else if ((current_byte2 = $B9) and (current_byte3 = $A3)) then // Small Hyphen-Minus
            begin
              current_value := $2D;
              output_file_stream.WriteBuffer(current_value, 1);
            end

            else if ((current_byte2 = $BC) and (current_byte3 = $8D)) then // Fullwidth Hyphen-Minus
            begin
              current_value := $2D;
              output_file_stream.WriteBuffer(current_value, 1);
            end

            end // if (current_byte = $EF) then

            else // not hyphen
            begin

            for icharset := 0 to 18 - 1 do
            begin
              table_value := LithuanianCharSetTable[icharset + 108]; // UTF-8 first byte
              table_value2 := LithuanianCharSetTable[icharset + 126];
              // UTF-8 second byte
              if ((table_value = current_byte) and (table_value2 = current_byte2)) then
              begin
                current_value := LithuanianCharSetTable[icharset +  (18 * output_charset_select)];

                if (current_value = 0) then
                begin
                  //remove null characters from file
                end
                else
                begin
                  output_file_stream.WriteBuffer(current_value, 1);
                end;

                Inc(number_of_replaced_letters);
                break;

              end; //if ((table_value = current_byte) and (table_value2 = current_byte2)) then
            end; //for icharset := 0 to 18 - 1 do

            end // else not hyphen

          end //else if(current_byte > 127)
          else
          begin
            output_file_stream.WriteBuffer(input_file_data[i], 1);
          end;
          //-----------

        end;//for i := start_i to file_size - 2 do


      finally
        output_file_stream.Free;
      end;
    except
      MessageDlgEx('Error: Unable to write output file.', mtInformation, [mbOK], LithuanianConverterForm);
    end;

  end //if((input_charset_select := 7) and (output_charset_select < 6))  // input UTF-8 and output not UTF-8
  //======================
  else if ((input_charset_select < 6) and (output_charset_select < 6)) then // if not UTF-8
  begin

    for i := 0 to file_size - 1 do
    begin
      current_value := input_file_data[i];

      if (current_value = 0) then
      begin
        //remove null characters from file
      end
      else if (current_value > 127) then
      begin
        for icharset := 0 to 18 - 1 do
        begin
          table_value := LithuanianCharSetTable[icharset + (18 * input_charset_select)];
          if (table_value = current_value) then
          begin
            input_file_data[i] :=
              LithuanianCharSetTable[icharset + (18 * output_charset_select)];
            Inc(number_of_replaced_letters);
            break;
          end;
          //if (table_value := value)
        end;

      end;
      //else if(value > 127)

    end; // for i := 0 to file_size - 1 do


    try
      output_file_stream := TFileStreamUTF8.Create(output_filename_string, fmCreate or fmOpenWrite);
      try
        output_file_stream.WriteBuffer(Pointer(input_file_data)^, Length(input_file_data));
      finally
        output_file_stream.Free;
      end;
    except
      MessageDlgEx('Error: Unable to write output file.', mtInformation, [mbOK], LithuanianConverterForm);
    end;

  end//if (output_charset_select <> 6) // if not UTF-8
  else if ((output_charset_select = 6) or (output_charset_select = 7)) then // UTF-8
  begin
    //===============
    if ((input_charset_select = 6) and (output_charset_select = 7)) then
    begin

      try
        output_file_stream := TFileStreamUTF8.Create(output_filename_string, fmCreate or fmOpenWrite);
        try
          for i := 3 to file_size - 1 do
          begin
            current_value := input_file_data[i];

            if (current_value = 0) then
            begin
              //remove null characters from file
            end
            else
            begin
              //output_file_stream.WriteBuffer(input_file_data[i], 1);
              output_file_stream.WriteBuffer(current_value, 1);
            end;

          end; //for i := 3 to file_size - 1 do

        finally
          output_file_stream.Free;
        end;
      except
        MessageDlgEx('Error: Unable to write output file.', mtInformation, [mbOK], LithuanianConverterForm);
      end;

    end //if((input_charset_select = 6) and (output_charset_select = 7)) then

    else if ((input_charset_select = 7) and (output_charset_select = 6)) then
    begin

      try
        output_file_stream := TFileStreamUTF8.Create(output_filename_string, fmCreate or fmOpenWrite);
        try

          one_byte := 239;
          output_file_stream.WriteBuffer(one_byte, 1);
          one_byte := 187;
          output_file_stream.WriteBuffer(one_byte, 1);
          one_byte := 191;
          output_file_stream.WriteBuffer(one_byte, 1);


          for i := 0 to file_size - 1 do
          begin
            current_value := input_file_data[i];

            if (current_value = 0) then
            begin
              //remove null characters from file
            end
            else
            begin
              //output_file_stream.WriteBuffer(input_file_data[i], 1);
              output_file_stream.WriteBuffer(current_value, 1);
            end;

          end; //for i := 0 to file_size-1 do

        finally
          output_file_stream.Free;
        end;
      except
        MessageDlgEx('Error: Unable to write output file.', mtInformation, [mbOK], LithuanianConverterForm);
      end;

    end //===============
    else // non-UTF-8 into UTF-8
    begin

      if (input_charset_select <> 4) then // if not Latin-7/ISO-8859-13/Baltic-RIM/ANSI-1257 charset
      begin

        for i := 0 to file_size - 1 do
        begin
          current_value := input_file_data[i];

          if (current_value = 0) then
          begin
            //remove null characters from file
          end
          else if (current_value > 127) then
          begin

            for icharset := 0 to 18 - 1 do
            begin
              table_value := LithuanianCharSetTable[icharset + (18 * input_charset_select)];
              if (table_value = current_value) then
              begin
                //MessageDlgEx('original := ' + IntToStr((int)input_file_data[i]), mtInformation, [mbOK], LithuanianConverterForm);
                input_file_data[i] := LithuanianCharSetTable[icharset + (18 * 4)];
                //convert into Windows-1257 charset
                Inc(number_of_replaced_letters);
                //MessageDlgEx('replaced to := ' + IntToStr((int)input_file_data[i]), mtInformation, [mbOK], LithuanianConverterForm);
                //break;
              end; // if (table_value = current_value) then
            end; // for icharset := 0 to 18 - 1 do

          end;
          //else if(value > 127)

        end; // for i := 0 to file_size - 1 do


        //if(number_of_replaced_letters = 0)
        //begin

        //MessageDlgEx('Step #1. Warning: input charset was not detected in input file.', mtInformation, [mbOK], LithuanianConverterForm);
        //end;

        //else
        //begin
        //MessageDlgEx('Step #1. Conversion completed. Number of replaced letters: ' + IntToStr(number_of_replaced_letters), mtInformation, [mbOK], LithuanianConverterForm);
      end;  //  if (input_charset_select <> 4) then // if not Latin-7/ISO-8859-13/Baltic-RIM/ANSI-1257 charset


    try

      output_file_stream := TFileStreamUTF8.Create(output_filename_string, fmCreate or fmOpenWrite);
      try
        //   if((output_charset_select = 6) and (OutputFileComboBox.ItemIndex <> 2)) // if UTF-8 BOM  and not ABBYY eFormFiller format
        if (output_charset_select = 6) then // if UTF-8 BOM
        begin
          one_byte := 239;
          output_file_stream.WriteBuffer(one_byte, 1);
          one_byte := 187;
          output_file_stream.WriteBuffer(one_byte, 1);
          one_byte := 191;
          output_file_stream.WriteBuffer(one_byte, 1);
        end;

        for i := 0 to file_size - 1 do
        begin
          current_value := input_file_data[i];

          if (current_value = 0) then
          begin
            //remove null characters from file
          end
          else if (current_value < 128) then
          begin
            output_file_stream.WriteBuffer(input_file_data[i], 1);
          end
          else if (current_value = 192) then
          begin
            one_byte := $C4;
            output_file_stream.WriteBuffer(one_byte, 1);
            one_byte := $84;
            output_file_stream.WriteBuffer(one_byte, 1);
            Inc(number_of_replaced_letters);
          end
          else if (current_value = 193) then
          begin
            one_byte := $C4;
            output_file_stream.WriteBuffer(one_byte, 1);
            one_byte := $AE;
            output_file_stream.WriteBuffer(one_byte, 1);
            Inc(number_of_replaced_letters);
          end
          else if (current_value = 198) then
          begin
            one_byte := $C4;
            output_file_stream.WriteBuffer(one_byte, 1);
            one_byte := $98;
            output_file_stream.WriteBuffer(one_byte, 1);
            Inc(number_of_replaced_letters);
          end
          else if (current_value = 200) then
          begin
            one_byte := $C4;
            output_file_stream.WriteBuffer(one_byte, 1);
            one_byte := $8C;
            output_file_stream.WriteBuffer(one_byte, 1);
            Inc(number_of_replaced_letters);
          end
          else if (current_value = 203) then
          begin
            one_byte := $C4;
            output_file_stream.WriteBuffer(one_byte, 1);
            one_byte := $96;
            output_file_stream.WriteBuffer(one_byte, 1);
            Inc(number_of_replaced_letters);
          end
          else if (current_value = 208) then
          begin
            one_byte := $C5;
            output_file_stream.WriteBuffer(one_byte, 1);
            one_byte := $A0;
            output_file_stream.WriteBuffer(one_byte, 1);
            Inc(number_of_replaced_letters);
          end
          else if (current_value = 216) then
          begin
            one_byte := $C5;
            output_file_stream.WriteBuffer(one_byte, 1);
            one_byte := $B2;
            output_file_stream.WriteBuffer(one_byte, 1);
            Inc(number_of_replaced_letters);
          end
          else if (current_value = 219) then
          begin
            one_byte := $C5;
            output_file_stream.WriteBuffer(one_byte, 1);
            one_byte := $AA;
            output_file_stream.WriteBuffer(one_byte, 1);
            Inc(number_of_replaced_letters);
          end
          else if (current_value = 222) then
          begin
            one_byte := $C5;
            output_file_stream.WriteBuffer(one_byte, 1);
            one_byte := $BD;
            output_file_stream.WriteBuffer(one_byte, 1);
            Inc(number_of_replaced_letters);
          end
          else if (current_value = 224) then
          begin
            one_byte := $C4;
            output_file_stream.WriteBuffer(one_byte, 1);
            one_byte := $85;
            output_file_stream.WriteBuffer(one_byte, 1);
            Inc(number_of_replaced_letters);
          end
          else if (current_value = 225) then
          begin
            one_byte := $C4;
            output_file_stream.WriteBuffer(one_byte, 1);
            one_byte := $AF;
            output_file_stream.WriteBuffer(one_byte, 1);
            Inc(number_of_replaced_letters);
          end
          else if (current_value = 230) then
          begin
            one_byte := $C4;
            output_file_stream.WriteBuffer(one_byte, 1);
            one_byte := $99;
            output_file_stream.WriteBuffer(one_byte, 1);
            Inc(number_of_replaced_letters);
          end
          else if (current_value = 232) then
          begin
            one_byte := $C4;
            output_file_stream.WriteBuffer(one_byte, 1);
            one_byte := $8D;
            output_file_stream.WriteBuffer(one_byte, 1);
            Inc(number_of_replaced_letters);
          end
          else if (current_value = 235) then
          begin
            one_byte := $C4;
            output_file_stream.WriteBuffer(one_byte, 1);
            one_byte := $97;
            output_file_stream.WriteBuffer(one_byte, 1);
            Inc(number_of_replaced_letters);
          end
          else if (current_value = 240) then
          begin
            one_byte := $C5;
            output_file_stream.WriteBuffer(one_byte, 1);
            one_byte := $A1;
            output_file_stream.WriteBuffer(one_byte, 1);
            Inc(number_of_replaced_letters);
          end
          else if (current_value = 248) then
          begin
            one_byte := $C5;
            output_file_stream.WriteBuffer(one_byte, 1);
            one_byte := $B3;
            output_file_stream.WriteBuffer(one_byte, 1);
            Inc(number_of_replaced_letters);
          end
          else if (current_value = 251) then
          begin
            one_byte := $C5;
            output_file_stream.WriteBuffer(one_byte, 1);
            one_byte := $AB;
            output_file_stream.WriteBuffer(one_byte, 1);
            Inc(number_of_replaced_letters);
          end
          else if (current_value = 254) then
          begin
            one_byte := $C5;
            output_file_stream.WriteBuffer(one_byte, 1);
            one_byte := $BE;
            output_file_stream.WriteBuffer(one_byte, 1);
            Inc(number_of_replaced_letters);
          end
          else
          begin
            //output_file_stream.WriteBuffer(one_byte, 1); // remove non-lithuanian characters
          end;

        end; //for i := 0 to file_size-1 do

      finally
        output_file_stream.Free;
      end;
    except
      MessageDlgEx('Error: Unable to write output file.', mtInformation, [mbOK], LithuanianConverterForm);
    end;

    end; // else // non-UTF-8 into UTF-8


    if (LithuanianCharsetPageControl.ActivePage = TabSheetSingle) then
    begin
      OutputMemo.Lines.LoadFromFile(output_filename_string);
    end;

  end;//if ((output_charset_select = 6) or (output_charset_select = 7)) // UTF-8



  SetLength(input_file_data, 0);

end;


procedure TLithuanianConverterForm.FormShow(Sender: TObject);
begin

  Left := (Screen.Width - Width) div 2;
  Top := (Screen.Height - Height) div 2;

end;

procedure TLithuanianConverterForm.IncludeSubdirectoriesCheckBoxChange(
  Sender: TObject);
begin

  include_subdirectories := IncludeSubdirectoriesCheckBox.Checked;

end;


procedure TLithuanianConverterForm.InputFileFormatComboBoxChange(Sender: TObject);
begin

  input_charset_select := InputFileFormatComboBox.ItemIndex - 1;

end;


procedure TLithuanianConverterForm.OpenInputDirectoryButtonClick(Sender: TObject);
begin

  if (CompareStr(old_input_directory, '') = 0) then
  begin
    old_input_directory := exe_filepath;
  end;



  SelectDirectory('Select input directory', old_input_directory, input_directory);
  output_directory := input_directory + '_out';
  input_directory := IncludeTrailingPathDelimiter(input_directory);
  output_directory := IncludeTrailingPathDelimiter(output_directory);

  old_input_directory := input_directory;
  InputDirectoryNameEdit.Text := input_directory;

  old_output_directory := output_directory;
  OutputDirectoryNameEdit.Text := output_directory;

end;

procedure TLithuanianConverterForm.OpenInputFileButtonClick(Sender: TObject);
var
  open_dialog: TOpenDialog;
begin

  open_dialog := TOpenDialog.Create(nil);

  open_dialog.Filter := 'All files (*.*)|*.*|Text files (*.txt)|*.txt|XML files (*.xml)|*.xml';

  if (open_dialog.Execute()) then
  begin

    InputFileNameEdit.Text := open_dialog.FileName;
    OutputFileComboBoxChange(Self);

    CheckIfFileIsUTF8BOM(InputFileNameEdit.Text);

  end; //if ( open_dialog.Execute() )


  open_dialog.Free;

end;

procedure TLithuanianConverterForm.OpenOutputDirectoryButtonClick(Sender: TObject);
begin

  if (CompareStr(old_output_directory, '') = 0) then
  begin
    old_output_directory := exe_filepath;
  end;


  SelectDirectory('Select output directory', old_output_directory, output_directory);
  old_output_directory := output_directory;
  OutputDirectoryNameEdit.Text := output_directory;

end;

procedure TLithuanianConverterForm.OpenOutputFileButtonClick(Sender: TObject);
var
  save_dialog: TSaveDialog;
begin

  save_dialog := TSaveDialog.Create(nil);

  if (eFormFiller_enabled = True) then
  begin
    save_dialog.Filter := 'ffdata files (*.ffdata)|*.ffdata|XML files (*.xml)|*.xml|Text files (*.txt)|*.txt';
  end
  else
  begin
    save_dialog.Filter := 'Text files (*.txt)|*.txt';
  end;



  if (save_dialog.Execute()) then
  begin
    OutputFileNameEdit.Text := save_dialog.FileName;
  end;


  save_dialog.Free;

end;

procedure TLithuanianConverterForm.OpenOutputFileCheckBoxChange(Sender: TObject);
begin

  open_file_after_conversion_select := OpenOutputFileCheckBox.Checked;

end;

procedure TLithuanianConverterForm.OutputFileComboBoxChange(Sender: TObject);
var
  input_file_name: string;
  file_extension: string;
begin

  if (InputFileNameEdit.Text <> '') then
  begin

    if (OutputFileComboBox.ItemIndex = 0) then
    begin

      input_file_name := ExtractFileDir(InputFileNameEdit.Text) +  PathDelim + ExtractFileName(InputFileNameEdit.Text);
      file_extension := ExtractFileExt(InputFileNameEdit.Text);
      OutputFileNameEdit.Text := input_file_name.SubString(0, Length(InputFileNameEdit.Text) -  Length(file_extension)) + '_out' + file_extension;
    end
    else if (OutputFileComboBox.ItemIndex = 1) then
    begin
      input_file_name := ExtractFileDir(InputFileNameEdit.Text) + PathDelim + ExtractFileName(InputFileNameEdit.Text);
      file_extension := ExtractFileExt(InputFileNameEdit.Text);
      OutputFileNameEdit.Text :=
        input_file_name.SubString(0, Length(InputFileNameEdit.Text) - Length(file_extension)) + '_out.txt';
    end
    else if (OutputFileComboBox.ItemIndex = 2) then
    begin
      OutputFileNameEdit.Text := ChangeFileExt(InputFileNameEdit.Text, '.ffdata');
    end;

  end //if(InputFileNameEdit.Text <> '')
  else
  begin
    OutputFileNameEdit.Text := '';
  end;
  //========

  if (OutputFileComboBox.Items.Count = 3) then
  begin

    if (OutputFileComboBox.ItemIndex = 0) then
    begin

      if (output_charset_select = 7) then // UTF-8
      begin
        output_charset_select := 6;
      end;
      // UTF-8 BOM (byte order mark)
    end
    else if (OutputFileComboBox.ItemIndex = 1) then
    begin

      if (output_charset_select = 7) then // UTF-8
      begin
        output_charset_select := 6;
      end;
      // UTF-8 BOM (byte order mark)

    end
    else if (OutputFileComboBox.ItemIndex = 2) then
    begin
      output_charset_select := 7; // UTF-8
    end;

    OutputFileFormatComboBox.ItemIndex := output_charset_select;

  end; //if (OutputFileComboBox.Items.Count = 3) then



  autofilenaming_charset_select := OutputFileComboBox.ItemIndex;

end;

procedure TLithuanianConverterForm.OutputFileFormatComboBoxChange(Sender: TObject);
begin

  output_charset_select := OutputFileFormatComboBox.ItemIndex;

end;

procedure TLithuanianConverterForm.StartBatchConversionButtonClick(Sender: TObject);
var
  i, i2: integer;
  output_directory_temp: string;
begin

  stop_conversion := False;


  Cursor := crHourGlass;
  StartBatchConversionButton.Enabled := False;
  StopBatchConversionButton.Enabled := True;

  number_of_converted_files := 0;

  OutputMemo.Lines.Clear();
  ResultsMemo.Lines.Clear();


  if (InputDirectoryNameEdit.Text = '') then
  begin

    MessageDlgEx('Please select input directory.', mtInformation, [mbOK], LithuanianConverterForm);
    Cursor := crDefault;
    StopBatchConversionButton.Enabled := False;
    StartBatchConversionButton.Enabled := True;
    Exit;
  end; //if(InputDirectoryNameEdit.Text = '')



  if (DirectoryExistsUTF8(InputDirectoryNameEdit.Text)) then
  begin

    if (OutputDirectoryNameEdit.Text = '') then
    begin

      MessageDlgEx('Please select output directory.', mtInformation, [mbOK], LithuanianConverterForm);
      Cursor := crDefault;
      StopBatchConversionButton.Enabled := False;
      StartBatchConversionButton.Enabled := True;
      Exit;
    end; //if(OutputDirectoryNameEdit.Text = '')


    input_directory := InputDirectoryNameEdit.Text;
    output_directory := OutputDirectoryNameEdit.Text;

    input_directory_length := Length(input_directory);
    output_directory_length := Length(output_directory);


    if (not (DirectoryExistsUTF8(output_directory))) then
    begin
      try
        begin

          MessageDlgEx('Output directory' + LineEnding +  output_directory + LineEnding + 'does not exist.' + LineEnding + 'The directory will be created.', mtInformation, [mbOK], LithuanianConverterForm);

          ForceDirectories(output_directory);

          if (not (DirectoryExistsUTF8(output_directory))) then
          begin

            MessageDlgEx('Error: failed to create' + LineEnding + output_directory + LineEnding + 'directory.', mtInformation, [mbOK], LithuanianConverterForm);

            Cursor := crDefault;
            StopBatchConversionButton.Enabled := False;
            StartBatchConversionButton.Enabled := True;
            Exit;

          end;  //if(not (DirectoryExistsUTF8(output_directory))) then

        end;


      except
        begin

          // Catch the possible exceptions
          MessageDlgEx('Error: failed to create' + LineEnding + output_directory + LineEnding + 'directory.', mtInformation, [mbOK], LithuanianConverterForm);

          Cursor := crDefault;
          StopBatchConversionButton.Enabled := False;
          StartBatchConversionButton.Enabled := True;
          Exit;
        end;

      end; //try

    end;  //if( not (DirectoryExistsUTF8(output_directory))) then


    ConvertProgressLabel.Caption := '0 %';
    ConvertProgressBar.Position := 0;


    file_extensions_string := FileExtensionsEdit.Text;

    input_file_formats_stringlist := TStringList.Create;
    try
      input_file_formats_stringlist.Delimiter := ';';
      input_file_formats_stringlist.DelimitedText := file_extensions_string;


      for i := 0 to input_file_formats_stringlist.Count - 1 do
      begin
//        MessageDlgEx(input_file_formats_stringlist[i], mtInformation, [mbOK], LithuanianConverterForm);

        if (stop_conversion = True) then
        begin

          Cursor := crDefault;
          StopBatchConversionButton.Enabled := False;
          StartBatchConversionButton.Enabled := True;
          break;
        end;


        //input_file_list := FindAllFiles(input_directory, '*.pas;*.pp;*.p;*.inc', IncludeSubdirectoriesCheckBox.Checked); //find e.g. all pascal sourcefiles
        input_file_list := FindAllFiles(input_directory, '*.' + input_file_formats_stringlist[i], IncludeSubdirectoriesCheckBox.Checked);
        try
          for i2 := 0 to input_file_list.Count - 1 do
          begin

            if (stop_conversion = True) then
            begin
              Cursor := crDefault;
              StopBatchConversionButton.Enabled := False;
              StartBatchConversionButton.Enabled := True;
              break;
            end;

            //XXXX   CheckIfFileIsUTF8BOM(input_file_list[i2]);

            output_filename := output_directory + Copy(input_file_list[i2], input_directory_length + 1, Length(input_file_list[i2]) - input_directory_length);

//            MessageDlgEx(output_filename, mtInformation, [mbOK], LithuanianConverterForm);

            output_directory_temp := ExtractFilePath(output_filename);

            if (not (DirectoryExistsUTF8(output_directory_temp))) then
            begin
              ForceDirectories(output_directory_temp);
            end;


            ConvertLithuanianFile(input_file_list[i2], output_filename);
            Inc(number_of_converted_files);
            ConvertProgressBar.Position := Floor(100.0 * (i2 + 1) / input_file_list.Count);
            ConvertProgressLabel.Caption := 'Conversion progress: ' + FloatToStr(RoundTo(ConvertProgressBar.Position, 0)) + ' % of "' + input_file_formats_stringlist[i] + '" extension files';
            Application.ProcessMessages();

          end; //for i2 := 0 to input_file_list.Count - 1 do

        finally
          input_file_list.Free;
        end;

      end; // for i := 0 to input_file_formats_stringlist.Count - 1 do


    finally
      input_file_formats_stringlist.Free;
    end;

  end  //if(TDirectory.Exists(InputDirectoryNameEdit.Text))
  else
  begin
    MessageDlgEx('Error: input directory' + LineEnding + InputDirectoryNameEdit.Text + LineEnding + 'does not exist', mtInformation, [mbOK], LithuanianConverterForm);
  end; //else -. if(TDirectory.Exists(InputDirectoryNameEdit.Text))



  ResultsMemo.Lines.Add('Conversion completed.' + LineEnding + 'Number of converted files: ' + IntToStr(number_of_converted_files));
  ResultsMemo.SelStart := 0;
  ResultsMemo.SetFocus(); // needed to invoke scrolling to the top
  OpenInputDirectoryButton.SetFocus();

  Cursor := crDefault;
  StopBatchConversionButton.Enabled := False;
  StartBatchConversionButton.Enabled := True;

end;

procedure TLithuanianConverterForm.StopBatchConversionButtonClick(Sender: TObject);
begin

  stop_conversion := True;

  StopBatchConversionButton.Enabled := False;
  StartBatchConversionButton.Enabled := True;

end;

procedure TLithuanianConverterForm.FormCreate(Sender: TObject);
{$IFDEF DARWIN}
var
  AppMenu: TMenuItem;
{$ENDIF}

begin
  {$IFDEF DARWIN}
  AppMenu := TMenuItem.Create(Self); // Application menu
  AppMenu.Caption := #$EF#$A3#$BF;   // Unicode Apple logo char
  MainMenu.Items.Insert(0, AppMenu);
  {$ENDIF}


  LithuanianCharsetPageControl.ActivePage := TabSheetSingle;

  file_extensions_string_default := 'txt;xml;c;h';


  input_directory := '';
  old_input_directory := '';
  output_directory := '';
  old_output_directory := '';

  file_extensions_string := '';
  stop_conversion := False;


  OutputMemo.Lines.Clear();
  ResultsMemo.Lines.Clear();


  {$IF Defined (DARWIN)}// macOS
  ini_filepath := GetUserDir();
  ini_filepath := ini_filepath +
    'Library/Application Support/Cognaxon/LithuanianCharSetConverter/';

  if (ForceDirectories(ini_filepath)) then
  begin
    ini_filename := ini_filepath + 'LithuanianCharSetConverter.ini';
  end;
  {$ELSEIF Defined (WINDOWS)}

  // https://www.freepascal.org/docs-html/rtl/sysutils/getappconfigdir.html
  //  C:\Users\mantas\AppData\Local\Cognaxon\LithuanianCharSetConverter\LithuanianCharSetConverter.ini


  ini_filename := ExtractFilePath(ParamStr(0)) + 'LithuanianCharSetConverter.ini';

  if (not (FileExistsUTF8(ini_filename))) then
  begin

  //ini_filepath := GetAppConfigDir(true);
  ini_filepath := GetAppConfigDir(False);
  ini_filepath := ExtractFilePath(ExcludeTrailingPathDelimiter(ini_filepath));
  //parentDirPath


  ini_filepath := ini_filepath + 'Cognaxon\LithuanianCharSetConverter\';

  if (ForceDirectories(ini_filepath)) then
  begin
    ini_filename := ini_filepath + 'LithuanianCharSetConverter.ini';
  end;

  end; //if (not (FileExistsUTF8(ini_filepath))) then

  {$ELSE}
//  ini_filename := ChangeFileExt(ParamStr(0), '.ini');
  ini_filename := ExtractFilePath(ParamStr(0)) + 'LithuanianCharSetConverter.ini';
  {$ENDIF}


  //ini_filename := AppendPathDelim(GetUserDir + 'Documents');

  try
    ProgramIniFile := TIniFile.Create(ini_filename);

    input_charset_select := ProgramIniFile.ReadInteger('LithuanianCharSetConverter', 'input_charset_select', -1); // Autodetect Lithuanian charset
    output_charset_select := ProgramIniFile.ReadInteger('LithuanianCharSetConverter', 'output_charset_select', 6); // UTF-8 BOM
    autofilenaming_charset_select := ProgramIniFile.ReadInteger('LithuanianCharSetConverter', 'autofilenaming_charset_select', 0);
    open_file_after_conversion_select := ProgramIniFile.ReadBool('LithuanianCharSetConverter', 'open_file_after_conversion_select', False);

    file_extensions_string := ProgramIniFile.ReadString('LithuanianCharSetConverter', 'file_extensions', file_extensions_string_default);

    input_directory := ProgramIniFile.ReadString('LithuanianCharSetConverter', 'input_directory', AppendPathDelim(GetUserDir + 'Documents'));
    output_directory := ProgramIniFile.ReadString('LithuanianCharSetConverter', 'output_directory', AppendPathDelim(GetUserDir + 'Documents'));

    include_subdirectories := ProgramIniFile.ReadBool('LithuanianCharSetConverter', 'include_subdirectories', False);

    eFormFiller_enabled := ProgramIniFile.ReadBool('LithuanianCharSetConverter', 'eFormFiller', False);

  finally
    ProgramIniFile.Free;
  end;



  FileExtensionsEdit.Text := file_extensions_string;

  IncludeSubdirectoriesCheckBox.Checked := include_subdirectories;




  InputFileFormatComboBox.Clear();
  InputFileFormatComboBox.Items.Add('Autodetect Lithuanian charset');  // index starts from 0
  InputFileFormatComboBox.Items.Add('ASCII');
  InputFileFormatComboBox.Items.Add('772 / Lithuanian Standard LST 1284:1993 (Lithuanian and Russian characters) ; 774 / Lithuanian Standard LST 1283:1993 (Lithuanian and English characters) ; 775 (Microsoft)');
  InputFileFormatComboBox.Items.Add('770 / IBM Baltic / Lithuanian Standard RST 1095-89');
  InputFileFormatComboBox.Items.Add('771 / KBL / Baltic Amadeus (Lithuanian and Russian characters) ; 773 Lithuanian (mix of 771 and 775)');
  InputFileFormatComboBox.Items.Add('Windows-1257 / IBM Baltic RIM ; Latin-7 / ISO-8859-13');
  InputFileFormatComboBox.Items.Add('Latin-4 / ISO-8859-4 ; Latin-6 / ISO-8859-10');
  InputFileFormatComboBox.Items.Add('UTF-8 BOM (byte order mark)');
  InputFileFormatComboBox.Items.Add('UTF-8');


  if (eFormFiller_enabled = True) then
  begin
    input_charset_select := 3;
  end;

  InputFileFormatComboBox.OnChange := nil;
  InputFileFormatComboBox.ItemIndex := input_charset_select + 1;
  InputFileFormatComboBox.OnChange := @InputFileFormatComboBoxChange;



  OutputFileFormatComboBox.Clear();
  OutputFileFormatComboBox.Items.Add('ASCII');
  OutputFileFormatComboBox.Items.Add('772 / Lithuanian Standard LST 1284:1993 (Lithuanian and Russian characters) ; 774 / Lithuanian Standard LST 1283:1993 (Lithuanian and English characters) ; 775 (Microsoft)');
  OutputFileFormatComboBox.Items.Add('770 / IBM Baltic / Lithuanian Standard RST 1095-89');
  OutputFileFormatComboBox.Items.Add('771 / KBL / Baltic Amadeus (Lithuanian and Russian characters) ; 773 Lithuanian (mix of 771 and 775)');
  OutputFileFormatComboBox.Items.Add('Windows-1257 / IBM Baltic RIM ; Latin-7 / ISO-8859-13');
  OutputFileFormatComboBox.Items.Add('Latin-4 / ISO-8859-4 ; Latin-6 / ISO-8859-10');
  OutputFileFormatComboBox.Items.Add('UTF-8 BOM (byte order mark)');
  OutputFileFormatComboBox.Items.Add('UTF-8');


  if (eFormFiller_enabled = True) then
  begin
    output_charset_select := 6;  // UTF-8 BOM
  end;

  OutputFileFormatComboBox.OnChange := nil;
  OutputFileFormatComboBox.ItemIndex := output_charset_select;
  OutputFileFormatComboBox.OnChange := @OutputFileFormatComboBoxChange;



  input_file_is_utf8_format := False;
  OutputFileComboBox.Clear();
  OutputFileComboBox.Items.Add('filename + _out + original file extention');
  OutputFileComboBox.Items.Add('filename + _out.txt');

  if (eFormFiller_enabled = True) then
  begin
    OutputFileComboBox.Items.Add('filename + .ffdata (ABBYY eFormFiller format)');
    OutputFileComboBox.ItemIndex := 2;
  end
  else
  begin
    OutputFileComboBox.ItemIndex := autofilenaming_charset_select;
  end;

  OutputFileComboBox.OnChange(Self);

  OpenOutputFileCheckBox.Checked := open_file_after_conversion_select;


  OutputFileNameEdit.Text := '';

  SetLength(charset_statistics_of_input_file, 256);




  AboutMemo.Lines.Clear();

  FileVerInfo := TFileVersionInfo.Create(nil);

  try
    FileVerInfo.ReadFileInfo;
    //AboutMemo.Lines.Add('Lithuanian charset converter v. ' + FileVerInfo.VersionStrings.Values['FileVersion']);
    AboutMemo.Lines.Add('Lithuanian charset converter v. ' + FileVerInfo.VersionStrings.Values['ProductVersion']);
  finally
    FileVerInfo.Free;
  end;

  AboutMemo.Lines.Add('');
  AboutMemo.Lines.Add('“Lithuanian charset converter” is free open-source cross-platform software made by Cognaxon');
  AboutMemo.Lines.Add('http://cognaxon.com');
  AboutMemo.Lines.Add('');
  AboutMemo.Lines.Add('Before UTF-8 character encoding adoption, Lithuanians had used different character encodings which were incompatible between each other.');
  AboutMemo.Lines.Add('“Lithuanian charset converter” converts between legacy character encodings and modern UTF-8.');

  AboutMemo.Lines.Add('');
  AboutMemo.Lines.Add('“Lithuanian charset converter” converts between:');
  AboutMemo.Lines.Add('• ASCII;');
  AboutMemo.Lines.Add('• 772/774/775;');
  AboutMemo.Lines.Add('• 770 (IBM Baltic);');
  AboutMemo.Lines.Add('• 771/773/KBL;');
  AboutMemo.Lines.Add('• Windows-1257/Baltic-RIM/Latin-7/ISO-8859-13;');
  AboutMemo.Lines.Add('• Latin-4/ISO-8859-4/Latin-6/ISO-8859-10;');
  AboutMemo.Lines.Add('• UTF-8;');
  AboutMemo.Lines.Add('• UTF-8 BOM.');
  AboutMemo.Lines.Add('');

  AboutMemo.Lines.Add('Online version of "Lithuanian charset converter" (HTML5 + JavaScript) is available at the address:');
  AboutMemo.Lines.Add('http://cognaxon.com/downloads/LithuanianCharSetConverter/online/Lithuanian_charset_converter.html');
  AboutMemo.Lines.Add('');
  AboutMemo.Lines.Add('Last updated: 2021-07-04');
  AboutMemo.Lines.Add('');
  AboutMemo.Lines.Add('INI file location:');
  AboutMemo.Lines.Add(ini_filename);





end;

procedure TLithuanianConverterForm.FormClose(Sender: TObject;
  var CloseAction: TCloseAction);
var
  ini_int_result: integer;
  ini_bool_result: boolean;
  ini_string_result: string;
begin


  input_charset_select := InputFileFormatComboBox.ItemIndex - 1;
  output_charset_select := OutputFileFormatComboBox.ItemIndex;


  try
    ProgramIniFile := TIniFile.Create(ini_filename);


    ini_int_result := ProgramIniFile.ReadInteger('LithuanianCharSetConverter', 'input_charset_select', -1); // Autodetect Lithuanian charset
    if (input_charset_select <> ini_int_result) then
    begin
      ProgramIniFile.WriteInteger('LithuanianCharSetConverter', 'input_charset_select', input_charset_select);
    end;


    ini_int_result := ProgramIniFile.ReadInteger('LithuanianCharSetConverter', 'output_charset_select', 6); // UTF-8 BOM
    if (output_charset_select <> ini_int_result) then
    begin
      ProgramIniFile.WriteInteger('LithuanianCharSetConverter', 'output_charset_select', output_charset_select);
    end;


    ini_int_result := ProgramIniFile.ReadInteger('LithuanianCharSetConverter', 'autofilenaming_charset_select', 0);
    if (autofilenaming_charset_select <> ini_int_result) then
    begin
      ProgramIniFile.WriteInteger('LithuanianCharSetConverter', 'autofilenaming_charset_select', autofilenaming_charset_select);
    end;


    ini_bool_result := ProgramIniFile.ReadBool('LithuanianCharSetConverter', 'open_file_after_conversion_select', False);
    if (open_file_after_conversion_select <> ini_bool_result) then
    begin
      ProgramIniFile.WriteBool('LithuanianCharSetConverter', 'open_file_after_conversion_select', open_file_after_conversion_select);
    end;


    ini_string_result := ProgramIniFile.ReadString('LithuanianCharSetConverter', 'file_extensions', file_extensions_string_default);

    file_extensions_string := FileExtensionsEdit.Text;

    if (CompareStr(file_extensions_string, ini_string_result) <> 0) then
    begin
      ProgramIniFile.WriteString('LithuanianCharSetConverter', 'file_extensions', file_extensions_string);
    end;



    include_subdirectories := IncludeSubdirectoriesCheckBox.Checked;
    ini_bool_result := ProgramIniFile.ReadBool('LithuanianCharSetConverter', 'include_subdirectories', False);
    if (include_subdirectories <> ini_bool_result) then
    begin
      ProgramIniFile.WriteBool('LithuanianCharSetConverter', 'include_subdirectories', include_subdirectories);
    end;


    //  ini_bool_result := ProgramIniFile.ReadBool( 'LithuanianCharSetConverter', 'eFormFiller', false);
    //  if(eFormFiller_enabled <> ini_bool_result) then
    //  begin
    //  ProgramIniFile.WriteBool( 'LithuanianCharSetConverter', 'eFormFiller', eFormFiller_enabled);
    //  end;


  finally
    ProgramIniFile.Free;
  end;


  SetLength(charset_statistics_of_input_file, 0);

end;

procedure TLithuanianConverterForm.ConvertSingleFileButtonClick(Sender: TObject);
begin

  if (InputFileNameEdit.Text = '') then
  begin
    MessageDlgEx('Please select input file.', mtInformation, [mbOK], LithuanianConverterForm);
    Exit;
  end;


  if not FileExistsUTF8(InputFileNameEdit.Text) then
  begin
    MessageDlgEx('Input file does not exist.' + LineEnding + 'Please select another input file.', mtInformation, [mbOK], LithuanianConverterForm);
    Exit;
  end;


  ConvertSingleFileButton.Enabled := False;
  OpenInputFileButton.Enabled := False;
  OpenOutputFileButton.Enabled := False;
  Cursor := crHourGlass;


  ConvertLithuanianFile(InputFileNameEdit.Text, OutputFileNameEdit.Text);


  ConvertSingleFileButton.Enabled := True;
  OpenInputFileButton.Enabled := True;
  OpenOutputFileButton.Enabled := True;
  Cursor := crDefault;


  if (input_charset_select = output_charset_select) then
  begin
    ResultsMemo.Lines.Add('Input and output charsets are the same. Conversion skipped.');
    ResultsMemo.SelStart := 0;
    ResultsMemo.SetFocus(); // needed to invoke scrolling to the top
    OpenInputFileButton.SetFocus();
    Exit;
  end;


  if (number_of_replaced_letters = 0) then
  begin
    //MessageDlgEx('Warning: input charset was not detected in input file.');
  end
  else
  begin
    if ((input_charset_select <> 4) and (output_charset_select > 5)) then
    begin
      ResultsMemo.Lines.Add('Conversion completed.' + LineEnding + 'Number of replaced letters: ' + IntToStr(number_of_replaced_letters div 2));
    end
    else
    begin
      ResultsMemo.Lines.Add('Conversion completed.' + LineEnding + 'Number of replaced letters: ' + IntToStr(number_of_replaced_letters));
    end;
  end;

  ResultsMemo.SelStart := 0;
  ResultsMemo.SetFocus(); // needed to invoke scrolling to the top
  OpenInputFileButton.SetFocus();

  if (OpenOutputFileCheckBox.Checked = True) then
  begin
    OpenDocument(OutputFileNameEdit.Text);
  end;

end;

end.
